unit Unit1;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Forms, Controls, Graphics, Dialogs, ExtCtrls, StdCtrls;

type

  { TForm1 }

  TForm1 = class(TForm)
    Button1: TButton;
    ColorButton1: TColorButton;
    Image1: TImage;
    Image2: TImage;
    Image3: TImage;
    Label1: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    procedure Button1Click(Sender: TObject);
    procedure FormCreate(Sender: TObject);
  private

  public

  end;

var
  Form1: TForm1;

implementation

{$R *.lfm}

uses
  LCLType, fpimage, intfgraphics;

{ TForm1 }

procedure ReplaceColorInBitmap(ABitmap: TBitmap; AColor, ANewColor: TColor);
var
  intfImg: TLazIntfImage;
  clr, newclr: TFPColor;
  imgHandle, imgMaskHandle: HBITMAP;
  i, j: Integer;
begin
  intfimg := ABitmap.CreateIntfImage;
  try
    clr := TColorToFPColor(AColor);
    newclr := TColorToFPColor(ANewColor);
    for i:=0 to intfimg.Width-1 do
      for j := 0 to intfimg.Height - 1 do
        if intfimg.Colors[i, j] = clr then
          intfimg.Colors[i, j] := newclr;
    intfimg.CreateBitmaps(imgHandle, imgMaskHandle, false);
    ABitmap.Handle := imgHandle;
    ABitmap.MaskHandle := imgMaskHandle;
  finally
    intfimg.Free;
  end;
end;

procedure ReplaceColorByFloodFill(ABitmap: TBitmap; AColor, ANewColor: TColor);
type
  TFillItem = record x,y: Integer; Color: TColor; end;
const
  BLOCKSIZE = 100;
var
  intfimg: TLazIntfImage;
  clr, c: TFPColor;
  i, j: Integer;
  n: Integer;
  Pts: array of TFillItem;
  found: Boolean;
begin
  n := 0;
  intfimg := ABitmap.CreateIntfImage;
  try
    clr := TColorToFPColor(AColor);
    found := false;
    for j:=0 to intfimg.Height - 1 do
      for i := 0 to intfimg.Width - 1 do begin
        c := intfimg.Colors[i, j];
        if c <> clr then
          found := false
        else if not found then begin
          if n mod BLOCKSIZE = 0 then
            SetLength(pts, Length(pts) + BLOCKSIZE);
          pts[n].X := i;
          pts[n].Y := j;
          pts[n].Color := FPColorToTColor(c);
          inc(n);
          found := true;
        end;
      end;
    SetLength(pts, n);
  finally
    intfimg.Free;
  end;

  for i:=0 to High(pts) do begin
    ABitmap.Canvas.Brush.Color := ANewColor;
    ABitmap.Canvas.FloodFill(Pts[i].X, Pts[i].Y, Pts[i].Color, fsSurface);
  end;
end;

procedure TForm1.Button1Click(Sender: TObject);
const
  N = 300;
var
  bmp: TBitmap;
  i: Integer;
  t0, t1, t2: TDateTime;
begin
  t0 := Now;
  for i:=1 to N do begin
    bmp := TBitmap.Create;
    try
      bmp.Assign(Image1.Picture.Bitmap);
      Image2.Picture.Bitmap.Assign(bmp);
    finally
      bmp.Free;
    end;
  end;
  t0 := Now - t0;
  Label1.Caption := 'Overhead: ' + FormatDateTime('s.zzz" sec"', t0);

  t1 := Now;
  for i:=1 to N do begin
    bmp := TBitmap.Create;
    try
      bmp.Assign(Image1.Picture.Bitmap);
      ReplaceColorInBitmap(bmp, clBlue, ColorButton1.ButtonColor);
      Image2.Picture.Bitmap.Assign(bmp);
    finally
      bmp.Free;
    end;
  end;
  t1 := Now - t1;
  Image2.Repaint;
  Label2.Caption := 'ReplaceColorInBitmap: ' + FormatDateTime('s.zzz" sec"', t1);

  t2 := Now;
  for i:=1 to N do begin
    bmp := TBitmap.Create;
    try
      bmp.Assign(Image1.Picture.Bitmap);
      ReplaceColorByFloodFill(bmp, clBlue, ColorButton1.ButtonColor);
      Image3.Picture.Bitmap.Assign(bmp);
    finally
      bmp.Free;
    end;
  end;
  t2 := Now - t2;
  Image3.Invalidate;

  Label3.Caption :=
    'ReplaceColorByFloodFill: ' + FormatDateTime('s.zzz', t2);
end;

procedure TForm1.FormCreate(Sender: TObject);
const
  COLORS: array[0..4] of TColor = (
    clBlue, clRed, clYellow, clMaroon, clGray);
//    clSilver, clOlive, clLime, clNavy, clGreen);
var
  bmp: TBitmap;
  i, j: Integer;
  P: array of TPoint;
begin
  RandSeed := 0;

  bmp := TBitmap.Create;
  try
    bmp.SetSize(Image1.Width, Image1.Height);
    bmp.Canvas.Brush.Color := clWhite;
    bmp.Canvas.FillRect(0, 0, bmp.Width, bmp.Height);
    for i:= Low(COLORS) to High(COLORS) do begin
      bmp.Canvas.Brush.Color := COLORS[i];
      if i = 0 then
        bmp.Canvas.Rectangle(40, 40, bmp.Width - 40, bmp.Height - 40)
      else begin
        SetLength(P, 3 + random(10));
        for j := 0 to High(P) do
          P[j] := Point(random(bmp.Width), random(bmp.Height));
        bmp.Canvas.Polygon(P);
      end;
    end;
    Image1.Picture.Assign(bmp);
  finally
    bmp.Free;
  end;
end;

end.

